import unittest
from datetime import datetime, timedelta

from django.conf import settings
from django.test import TestCase

import balancer
from balancer.routers import RandomRouter, RoundRobinRouter, \
                             WeightedRandomRouter, \
                             WeightedMasterSlaveRouter, \
                             RoundRobinMasterSlaveRouter, \
                             PinningWMSRouter, PinningRRMSRouter
from balancer.mixins import PinningMixin
from balancer.middleware import PINNING_KEY, PINNING_SECONDS, \
                                PinningSessionMiddleware, \
                                PinningCookieMiddleware

class BalancerTestCase(TestCase):

    def setUp(self):
        self.original_databases = settings.DATABASES
        settings.DATABASES = balancer.TEST_DATABASES

        self.original_master = getattr(settings, 'MASTER_DATABASE', None)
        settings.MASTER_DATABASE = balancer.TEST_MASTER_DATABASE

        self.original_pool = getattr(settings, 'DATABASE_POOL', None)
        settings.DATABASE_POOL = balancer.TEST_DATABASE_POOL

        class MockObj(object):
            class _state:
                db = None

        self.obj1 = MockObj()
        self.obj2 = MockObj()

    def tearDown(self):
        settings.DATABASES = self.original_databases
        settings.MASTER_DATABASE = self.original_master
        settings.DATABASE_POOL = self.original_pool


class RandomRouterTestCase(BalancerTestCase):

    def setUp(self):
        super(RandomRouterTestCase, self).setUp()
        self.router = RandomRouter()

    def test_random_db_selection(self):
        """Simple test to make sure that random database selection works."""
        for i in range(10):
            self.assertIn(self.router.get_random_db(),
                          settings.DATABASE_POOL.keys())

    def test_relations(self):
        """Relations should only be allowed for databases in the pool."""
        self.obj1._state.db = 'default'
        self.obj2._state.db = 'other'
        self.assertTrue(self.router.allow_relation(self.obj1, self.obj2))

        self.obj1._state.db = 'other'
        self.obj2._state.db = 'utility'
        self.assertFalse(self.router.allow_relation(self.obj1, self.obj2))


class RoundRobinRouterTestCase(BalancerTestCase):

    def setUp(self):
        super(RoundRobinRouterTestCase, self).setUp()
        settings.DATABASE_POOL = ['default', 'other', 'utility']
        self.router = RoundRobinRouter()
    
    def test_sequential_db_selection(self):
        """Databases should cycle in order."""
        for i in range(10):
            self.assertEqual(self.router.get_next_db(), self.router.pool[0])
            self.assertEqual(self.router.get_next_db(), self.router.pool[1])
            self.assertEqual(self.router.get_next_db(), self.router.pool[2])


class WeightedRandomRouterTestCase(BalancerTestCase):

    def setUp(self):
        super(WeightedRandomRouterTestCase, self).setUp()
        self.router = WeightedRandomRouter()

    def test_weighted_db_selection(self):
        """
        Make sure that the weights are being applied correctly by checking to
        see if the rate that 'default' is selected is within 0.15 of the target
        rate.
        """
        def check_rate(target):
            hits = {'default': 0, 'other': 0}
            for i in range(1000):
                hits[self.router.get_random_db()] += 1
            rate = round(float(hits['default']) / float(hits['other']), 2)

            self.assertTrue((target - 0.15) <= rate <= (target + 0.15),
                            "The 'default' rate of %s was not close enough to "
                            "the target rate." % rate)

        # The initial target rate is 0.5, because 'default' has a weight of 1
        # and 'other' has a rate of 2 - 'default' should be selected roughly
        # half as much as 'other'.
        check_rate(target=0.5)

        settings.DATABASE_POOL = {
            'default': 1,
            'other': 4,
        }
        # Reinitialize the router with new weights
        self.router = WeightedRandomRouter()
        check_rate(target=0.25)


class MasterSlaveTestMixin(object):
    """A mixin for testing routers that use the MasterSlaveMixin."""
    
    def test_writes(self):
        """Writes should always go to master."""
        self.assertEqual(self.router.db_for_write(self.obj1), 'default')

    def test_relations(self):
        """
        Relations should be allowed for databases in the pool and the master.
        """
        settings.DATABASE_POOL = {
            'other': 1,
            'utility': 1,
        }
        self.router = WeightedRandomRouter()

        # Even though default isn't in the database pool, it is the master so
        # the relation should be allowed.
        self.obj1._state.db = 'default'
        self.obj2._state.db = 'other'
        self.assertTrue(self.router.allow_relation(self.obj1, self.obj2))


class WMSRouterTestCase(MasterSlaveTestMixin, BalancerTestCase):
    """Tests for the WeightedMasterSlaveRouter."""

    def setUp(self):
        super(WMSRouterTestCase, self).setUp()
        self.router = WeightedMasterSlaveRouter()


class RRMSRouterTestCase(MasterSlaveTestMixin, BalancerTestCase):
    """Tests for the RoundRobinMasterSlaveRouter."""
    
    def setUp(self):
        super(RRMSRouterTestCase, self).setUp()
        self.router = RoundRobinMasterSlaveRouter()


class PinningRouterTestMixin(object):
    """A mixin for testing routers that use the pinning mixin."""

    def setUp(self):
        super(PinningRouterTestMixin, self).setUp()
        
        class MockRequest(object):
            session = {}
            COOKIES = []
        
        self.mock_request = MockRequest()
        
        class MockResponse(object):
            cookie = None
            
            def set_cookie(self, key, value, max_age):
                self.cookie = key
        
        self.mock_response = MockResponse()

    def test_pinning(self):
        # Check to make sure the 'other' database shows in in reads first
        success = False
        for i in range(100):
            db = self.router.db_for_read(self.obj1)
            if db == 'other':
                success = True
                break
        self.assertTrue(success, "The 'other' database was not offered.")
        
        # Simulate a write
        self.router.db_for_write(self.obj1)
        
        # Check to make sure that only the master database shows up in reads,
        # since the thread should now be pinned
        success = True
        for i in range(100):
            db = self.router.db_for_read(self.obj1)
            if db == 'other':
                success = False
                break
        self.assertTrue(success, "The 'other' database was offered in error.")
        
        PinningMixin.unpin_thread()
        PinningMixin.clear_db_write()
    
    def test_middleware(self):
        for middleware, vehicle in [(PinningSessionMiddleware(), 'session'),
                                    (PinningCookieMiddleware(), 'cookie')]:
            # The first request shouldn't pin the database
            middleware.process_request(self.mock_request)
            self.assertFalse(PinningMixin.is_pinned())
            
            # A simulated write should, however
            PinningMixin.set_db_write()
            
            # The response should set the session variable and clear the locals
            response = middleware.process_response(self.mock_request,
                                                   self.mock_response)
            self.assertFalse(PinningMixin.is_pinned())
            self.assertFalse(PinningMixin.db_was_written())
            if vehicle == 'session':
                self.assertTrue(
                    self.mock_request.session.get(PINNING_KEY, False)
                )
            else:
                self.assertEqual(response.cookie, PINNING_KEY)
                self.mock_request.COOKIES = [response.cookie]
            
            # The subsequent request should then pin the database
            middleware.process_request(self.mock_request)
            self.assertTrue(PinningMixin.is_pinned())
            
            PinningMixin.unpin_thread()
            
            if vehicle == 'session':
                # After the pinning period has expired, the request should no
                # longer pin the thread
                exp = timedelta(seconds=PINNING_SECONDS - 5)
                self.mock_request.session[PINNING_KEY] = datetime.now() - exp
                middleware.process_request(self.mock_request)
                self.assertFalse(PinningMixin.is_pinned())
                
                PinningMixin.unpin_thread()


class PinningWMSRouterTestCase(PinningRouterTestMixin, BalancerTestCase):
    
    def setUp(self):
        super(PinningWMSRouterTestCase, self).setUp()
        self.router = PinningWMSRouter()


class PinningRRMSRouterTestCase(PinningRouterTestMixin, BalancerTestCase):
    
    def setUp(self):
        super(PinningRRMSRouterTestCase, self).setUp()
        self.router = PinningRRMSRouter()



VERSION = (0, 3, 0, "a", 1) # following PEP 386
DEV_N = None # for PyPi releases, set this to None


def get_version(short=False):
    version = "%s.%s" % (VERSION[0], VERSION[1])
    if short:
        return version
    if VERSION[2]:
        version = "%s.%s" % (version, VERSION[2])
    if VERSION[3] != "f":
        version = "%s%s%s" % (version, VERSION[3], VERSION[4])
        if DEV_N:
            version = "%s.dev%s" % (version, DEV_N)
    return version

__version__ = get_version()

TEST_DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
    },
    'other': {
        'ENGINE': 'django.db.backends.sqlite3',
        'TEST_NAME': 'other_db',
    },
    'utility': {
        'ENGINE': 'django.db.backends.sqlite3',
        'TEST_NAME': 'utility_db',
    }
}
TEST_MASTER_DATABASE = 'default'
TEST_DATABASE_POOL = {
    'default': 1,
    'other': 2,
}



import threading

from django.conf import settings


_locals = threading.local()


class MasterSlaveMixin(object):
    """
    A mixin that randomly selects from a weighted pool of slave databases
    for read operations, but uses the default database for writes.
    """

    def __init__(self):
        super(MasterSlaveMixin, self).__init__()
        self.master = settings.MASTER_DATABASE

    def db_for_write(self, model, **hints):
        """Send all writes to the master"""
        return self.master

    def allow_relation(self, obj1, obj2, **hints):
        """
        Allow any relation between two objects in the slave pool or the master.
        """
        pool = self.pool + [self.master]
        if obj1._state.db in pool and obj2._state.db in pool:
            return True
        return None

    def allow_syncdb(self, db, model):
        """Only allow syncdb on the master"""
        return db == self.master


class PinningMixin(object):
    """
    A mixin that pins reads to the database defined in the MASTER_DATABASE
    setting for a pre-determined period of time after a write.  Requires the
    PinningRouterMiddleware.
    """
    
    def db_for_read(self, model, **hints):
        if PinningMixin.is_pinned():
            return settings.MASTER_DATABASE
        return super(PinningMixin, self).db_for_read(model, **hints)
    
    def db_for_write(self, model, **hints):
        PinningMixin.set_db_write()
        PinningMixin.pin_thread()
        return super(PinningMixin, self).db_for_write(model, **hints)
    
    
    @staticmethod
    def pin_thread():
        """
        Mark this thread as 'pinned', so that future reads will temporarily go
        to the master database for the current user.  
        """
        _locals.pinned = True
    
    @staticmethod
    def unpin_thread():
        """
        Clear the 'pinned' flag so that future reads are distributed normally.
        """
        if getattr(_locals, 'pinned', False):
            del _locals.pinned
    
    @staticmethod
    def is_pinned():
        """Check whether the current thread is pinned."""
        return getattr(_locals, 'pinned', False)
    
    @staticmethod
    def set_db_write():
        """Indicate that the database was written to."""
        _locals.db_write = True
    
    @staticmethod
    def clear_db_write():
        if getattr(_locals, 'db_write', False):
            del _locals.db_write
    
    @staticmethod
    def db_was_written():
        """Check whether a database write was performed."""
        return getattr(_locals, 'db_write', False)


import bisect
import itertools
import random
import warnings

from django.conf import settings

from balancer.mixins import MasterSlaveMixin, PinningMixin


class BasePoolRouter(object):
    """
    A base class for routers that use a pool of databases defined by the
    DATABASE_POOL setting.
    """
    
    def __init__(self):
        if isinstance(settings.DATABASE_POOL, dict):
            self.pool = settings.DATABASE_POOL.keys()
        else:
            self.pool = settings.DATABASE_POOL
    
    def allow_relation(self, obj1, obj2, **hints):
        """Allow any relation between two objects in the pool"""
        if obj1._state.db in self.pool and obj2._state.db in self.pool:
            return True
        return None

    def allow_syncdb(self, db, model):
        """Explicitly put all models on all databases"""
        return True


class RandomRouter(BasePoolRouter):
    """A router that randomly selects from a pool of databases."""

    def db_for_read(self, model, **hints):
        return self.get_random_db()

    def db_for_write(self, model, **hints):
        return self.get_random_db()

    def get_random_db(self):
        return random.choice(self.pool)


class WeightedRandomRouter(RandomRouter):
    """
    A router that randomly selects from a weighted pool of databases, useful
    for replication configurations where all nodes act as masters.
    """

    def __init__(self):
        self.pool = settings.DATABASE_POOL.keys()
        self.totals = []

        weights = settings.DATABASE_POOL.values()
        running_total = 0

        for w in weights:
            running_total += w
            self.totals.append(running_total)

    def get_random_db(self):
        """Use binary search to find the index of the database to use"""
        rnd = random.random() * self.totals[-1]
        pool_index = bisect.bisect_right(self.totals, rnd)
        return self.pool[pool_index]


class RoundRobinRouter(BasePoolRouter):
    """
    A router that cycles over a pool of databases in order, evenly distributing
    the load.
    """
    
    def __init__(self):
        super(RoundRobinRouter, self).__init__()
        
        # Shuffle the pool so the first database isn't slammed during startup.
        random.shuffle(self.pool)
        
        self.pool_cycle = itertools.cycle(self.pool)
    
    def db_for_read(self, model, **hints):
        return self.get_next_db()

    def db_for_write(self, model, **hints):
        return self.get_next_db()
    
    def get_next_db(self):
        return self.pool_cycle.next()


class WeightedMasterSlaveRouter(MasterSlaveMixin, WeightedRandomRouter):
    pass


class RoundRobinMasterSlaveRouter(MasterSlaveMixin, RoundRobinRouter):
    pass


class PinningWMSRouter(PinningMixin, WeightedMasterSlaveRouter):
    """A weighted master/slave router that uses the pinning mixin."""
    pass


class PinningMasterSlaveRouter(PinningWMSRouter):
    """An alias to PinningWMSRouter.  This will be removed in 0.4."""
    
    def __init__(self):
        warnings.warn("This router has been renamed to 'PinningWMSRouter', "
                      "and it will be removed in the next release.",
                      DeprecationWarning)
        super(PinningMasterSlaveRouter, self).__init__()


class PinningRRMSRouter(PinningMixin, RoundRobinMasterSlaveRouter):
    """A round-robin master/slave router that uses the pinning mixin."""
    pass



from datetime import datetime, timedelta

from django.conf import settings

from balancer.mixins import PinningMixin


# The name of the session variable or cookie used by the middleware
PINNING_KEY = getattr(settings, 'MASTER_PINNING_KEY', 'master_db_pinned')

# The number of seconds to direct reads to the master database after a write
PINNING_SECONDS = int(getattr(settings, 'MASTER_PINNING_SECONDS', 5))


class PinningSessionMiddleware(object):
    """
    Middleware to support the PinningMixin.  Sets a session variable if
    there was a database write, which will direct that user's subsequent reads
    to the master database.
    """
    
    def process_request(self, request):
        """
        Set the thread's pinning flag according to the presence of the session
        variable.
        """
        pinned_until = request.session.get(PINNING_KEY, False)
        if pinned_until and pinned_until > datetime.now():
            PinningMixin.pin_thread()
        
    def process_response(self, request, response):
        """
        If there was a write to the db, set the session variable to enable
        pinning.  If the variable already exists, the time will be reset.
        """
        if PinningMixin.db_was_written():
            pinned_until = datetime.now() + timedelta(seconds=PINNING_SECONDS)
            request.session[PINNING_KEY] = pinned_until
            PinningMixin.clear_db_write()
        PinningMixin.unpin_thread()
        return response


class PinningCookieMiddleware(object):
    """
    Middleware to support the PinningMixin.  Sets a cookie if there was a
    database write, which will direct that user's subsequent reads to the
    master database.
    """
    
    def process_request(self, request):
        """
        Set the thread's pinning flag according to the presence of the cookie.
        """
        if PINNING_KEY in request.COOKIES:
            PinningMixin.pin_thread()
    
    def process_response(self, request, response):
        """
        If there was a write to the db, set the cookie to enable pinning.  If
        the cookie already exists, the time will be reset.
        """
        if PinningMixin.db_was_written():
            response.set_cookie(PINNING_KEY, value='y',
                                max_age=PINNING_SECONDS)
            PinningMixin.clear_db_write()
        PinningMixin.unpin_thread()
        return response



from distutils.core import setup

VERSION = __import__('balancer').__version__

try:
    long_description = open('README', 'rt').read()
except IOError:
    long_description = ''

description = "A set of tools for using Django's multi-db feature to balance "
description += "database requests"

setup(
    name='django-balancer',
    version=VERSION,
    description=description,
    long_description = long_description,
    author='Brandon Konkle',
    author_email='brandon@brandonkonkle.com',
    license='License :: OSI Approved :: BSD License',
    url='http://github.com/bkonkle/django-balancer',
    packages=['balancer'],
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Development Status :: 3 - Alpha',
    ]
)
